// client-side main module

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#ifndef WIN32
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>
#endif
#include <signal.h>
#include "utils.h"
#include "commondefs.h"

#define WIN_WIDTH (450)  // window size
#define WIN_HEIGHT (350)

static int Nclient; // number of clients
static Gamestatus Gs;
static int GsCnt; // general-purpose counter
static struct {
    int hit;
    int try;
} Statistics; // statistics for data transfer optimization
static char Map[MAP_HEIGHT][MAP_WIDTH];
static int Smoke[MAX_SMOKE][3]; // position & lifetime of smokes
static int Stage;
static int Nenemy;
static int Nflag;
static int NflagLeft;
static int Highscore;
static int Winner;
static int Soc; // socket to server
static char RecvBuf[MAX_DATA_SIZE];
static char SendBuf[MAX_DATA_SIZE];
static Player P[MAX_CLIENT]; // data of all players
static Player *Me; // Me points to P[myid]
static int WinWidth = WIN_WIDTH;
static int WinHeight = WIN_HEIGHT;
static void (* unpack_func[NTAG])();
static void (* game_func[NGAME])();

// data transfer routines
static void init_vars(void);
static void init_session(void);
static void send_to_server(int size);
static void recv_from_server(void);
static void pack_keyinput(unsigned int key);
static void unpack_playerinfo(void);
static void unpack_playerinfo2(void);
static void unpack_stageinfo(void);
static void unpack_map(void);
static void unpack_smoke(void);
static void unpack_highscore(void);
static void unpack_gametitle(void);
static void unpack_gameready(void);
static void unpack_gamestart(void);
static void unpack_gamefinish(void);
static void unpack_gameresult(void);
static void unpack_gamefinalresult(void);
static void unpack_gameretire(void);
static void unpack_gameover(void);
static void unpack_gamequit(void);
static void unpack_gamepause(void);

// state machine
static void game_title(void);
static void game_ready(void);
static void game_onplay(void);
static void game_retire(void);
static void game_result(void);
static void game_pause(void);
static void game_finalresult(void);
static void game_finish(void);
static void game_over(void);

// OpenGL event handlers
static void myDisplayTitle(void);
static void myDisplayOnPlay(void);
static void myDisplayRetire(void);
static void myDisplayResult(void);
static void myDisplayFinalResult(void);
static void myKeyboard(unsigned char key, int x, int y);
static void myKeyboardUp(unsigned char key, int x, int y);
static void mySpecial(int key, int x, int y);
static void mySpecialUp(int key, int x, int y);
static void myTimer(int id);
static void myReshape(int w, int h);
static void kickoff_opengl(int argc, char *argv[]);
static void display_play_field(void);
static void display_whole_play_field(void);

// character drawers
static void putCar(Player *p);
static void putMyCar(void);
static void putWall0(void);
static void putWall(int mx, int my);
static void putWallBorder0(double angle);
static void putWallBorder(int mx, int my);
static void putRoad(double x, double y);
static void putFlag(double x, double y);
static void putLuckyFlag(double x, double y);
static void putSpecialFlag(double x, double y);
static void putRock(double x, double y);
static void putFuel(void);
static void putRader(void);
static void putBackground(void);
static void putInfo(void);
static void putScore(void);
static void putMyCarAt(double x, double y);
static void putMyCar0(void);
static void putFlagAt(double x, double y, double b, int withcross);
static void putFlag0(void);
static void putSmoke(double x, double y, int lifetime);
static void putMessages(void);
static void putTitle(void);
static void putResult(void);
static void putFinalResult(void);

// utilities
static int p2id(Player *p);

// just for debugging
static void print_map(void);

int
main(int argc, char *argv[])
{
    char hostname[128];

    printf("server name or IP  [127.0.0.1]: ");
    fgets(hostname, sizeof(hostname), stdin);
    if (strlen(hostname) < 2) {
	sprintf(hostname, "127.0.0.1");
    }

    chop_newline(hostname, sizeof(hostname));

    Soc = setup_client(hostname, PORT);
    if (Soc == -1) {
	exit(1);
    }
    init_vars();
    init_session();
    kickoff_opengl(argc, argv);
    die(0, 1, &Soc);
}

static void
init_vars(void)
{
    unpack_func[TAG_PLAYERINFO] = unpack_playerinfo;
    unpack_func[TAG_PLAYERINFO2] = unpack_playerinfo2;
    unpack_func[TAG_STAGEINFO] = unpack_stageinfo;
    unpack_func[TAG_MAP] = unpack_map;
    unpack_func[TAG_SMOKE] = unpack_smoke;
    unpack_func[TAG_HIGHSCORE] = unpack_highscore;
    unpack_func[TAG_GAMETITLE] = unpack_gametitle;
    unpack_func[TAG_GAMEREADY] = unpack_gameready;
    unpack_func[TAG_GAMESTART] = unpack_gamestart;
    unpack_func[TAG_GAMEFINISH] = unpack_gamefinish;
    unpack_func[TAG_GAMERESULT] = unpack_gameresult;
    unpack_func[TAG_GAMEFINALRESULT] = unpack_gamefinalresult;
    unpack_func[TAG_GAMERETIRE] = unpack_gameretire;
    unpack_func[TAG_GAMEOVER] = unpack_gameover;
    unpack_func[TAG_GAMEQUIT] = unpack_gamequit;
    unpack_func[TAG_GAMEPAUSE] = unpack_gamepause;

    game_func[GAME_TITLE] = game_title;
    game_func[GAME_READY] = game_ready;
    game_func[GAME_ONPLAY] = game_onplay;
    game_func[GAME_FINISH] = game_finish;
    game_func[GAME_OVER] = game_over;
    game_func[GAME_RETIRE] = game_retire;
    game_func[GAME_RESULT] = game_result;
    game_func[GAME_PAUSE] = game_pause;
    game_func[GAME_FINALRESULT] = game_finalresult;
}

static void
kickoff_opengl(int argc, char *argv[])
{
    char buf[32];

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGB);
    glutInitWindowSize(WIN_WIDTH, WIN_HEIGHT);
    sprintf(buf, "Rally-Z client (ID:%02d)", p2id(Me));
    glutCreateWindow(buf);
    glutDisplayFunc(myDisplayTitle);
    glutTimerFunc(10, myTimer, 0);
    glutKeyboardFunc(myKeyboard); 
    glutKeyboardUpFunc(myKeyboardUp); 
    glutSpecialFunc(mySpecial);
    glutSpecialUpFunc(mySpecialUp);
    glutIgnoreKeyRepeat(1); // nonzero:ignore repeat    zero:repeat
    glClearColor(0.0, 0.0, 0.3, 0.0);
    glutReshapeFunc(myReshape);
    glutMainLoop();
}

static void
init_session(void)
{
    int myid, len;
    char mesg[128];

    // login session
    recv(Soc, RecvBuf, sizeof(int), 0);
    len = ntohl(*(int *)RecvBuf);
    recv(Soc, RecvBuf, len, 0);
    printf(RecvBuf); // message from server: "your name: "
    fgets(mesg, MAX_NAMELEN, stdin);
    chop_newline(mesg, MAX_NAMELEN);
    len = strlen(mesg) + 1;
    *(int *)SendBuf = htonl(len);
    strcpy(SendBuf + sizeof(int), mesg);
    send(Soc, SendBuf, sizeof(int) + len, 0); /* send player name
						 (e.g. A.Kawai) to server */
    recv(Soc, RecvBuf, sizeof(int), 0);
    myid = ntohl(*(int *)RecvBuf);
    printf("your entry number is %d\n", myid);
    Me = P+myid;

    // wait until all clients are logged in....

    recv(Soc, RecvBuf, sizeof(int), 0);
    Nclient = ntohl(*(int *)RecvBuf);
    printf("number of players: %d\n", Nclient);

    recv(Soc, RecvBuf, sizeof(int), 0);
    len = ntohl(*(int *)RecvBuf);
    recv(Soc, RecvBuf, len, 0);
    printf(RecvBuf); // message from server: "Hello, ..."
    Gs = GAME_TITLE;
}

static void
pack_keyinput(unsigned int key)
{
    int size;

    ((int *)(SendBuf))[0] = htonl(TAG_KEYPRESSED);
    size = TAG_SIZE + sizeof(int);
    ((int *)(SendBuf))[1] = htonl(size);
    ((int *)(SendBuf))[2] = htonl(key);
    send_to_server(size);
}

static void
send_to_server(int size)
{
    send(Soc, SendBuf, size, 0);
}

static void
recv_from_server(void)
{
    int tag;
    long int size;

    Statistics.try++;

#if defined(WIN32)
    ioctlsocket(Soc, FIONREAD, &size);
    if (size < TAG_SIZE) return; // no data arrived yet.
#else
    size = recv(Soc, RecvBuf, TAG_SIZE, MSG_PEEK | MSG_DONTWAIT);
    if (size < 0) {
	if (errno == EAGAIN) { // no data arrived yet.
	    return;
	}
	else {
	    perror("recv_from_server EAGAIN");
	}
    }
    if (size < TAG_SIZE) return; // no data arrived yet.
#endif

    Statistics.hit++; // something in read queue

    recv(Soc, RecvBuf, TAG_SIZE, MSG_PEEK);
    tag = ntohl(((int *)RecvBuf)[0]);
    size = ntohl(((int *)RecvBuf)[1]);
    recv(Soc, RecvBuf, size, 0);

    //    fprintf(stderr, "tag:%d  size:%d\n", tag, size);

    if (tag < 0 || tag >= NTAG) {
	fprintf(stderr, "unknown tag: %d\n", tag);
    }
    (unpack_func[tag])();
}

static void
unpack_playerinfo(void)
{
    int i, off;
    static int cnt = 0;
    off = 2; // offset for tag
    for (i = 0; i < Nclient + Nenemy; i++) {
	P[i].x = ntohl(((int *)(RecvBuf))[off++]);
	P[i].y = ntohl(((int *)(RecvBuf))[off++]);
	P[i].dir = ntohl(((int *)(RecvBuf))[off++]);
	P[i].status = ntohl(((int *)(RecvBuf))[off++]);
	P[i].fuel = ntohl(((int *)(RecvBuf))[off++]);
    }
    if (sizeof(int) * off != TAG_SIZE+POFF * (Nclient + Nenemy)) {
	fprintf(stderr,
		"unpack_playerinfo off:%d  TAG_SIZE+POFF*(Nclient+Nenemy):%d\n"
		"something wrong. abort...\n",
		sizeof(int) * off, TAG_SIZE+POFF * (Nclient + Nenemy));
    }
    //    printf("cnt: %d x:%+03d  y:%+03d  dir:%3d\n",
    //    cnt++, Me->x, Me->y, Me->dir);
}

static void
unpack_playerinfo2(void)
{
    int i, off;
    static int cnt = 0;

    off = 2; // offset for tag
    for (i = 0; i < Nclient + Nenemy; i++) {
	P[i].x = ntohl(((int *)(RecvBuf))[off++]);
	P[i].y = ntohl(((int *)(RecvBuf))[off++]);
	P[i].dir = ntohl(((int *)(RecvBuf))[off++]);
	P[i].status = ntohl(((int *)(RecvBuf))[off++]);
	P[i].fuel = ntohl(((int *)(RecvBuf))[off++]);
	P[i].nflag = ntohl(((int *)(RecvBuf))[off++]);
	P[i].nlife = ntohl(((int *)(RecvBuf))[off++]);
	P[i].nwin = ntohl(((int *)(RecvBuf))[off++]);
	P[i].score = ntohl(((int *)(RecvBuf))[off++]);
	P[i].ntotalflag = ntohl(((int *)(RecvBuf))[off++]);
	P[i].ctrlstyle = ntohl(((int *)(RecvBuf))[off++]);
	memcpy(P[i].name, RecvBuf + off * sizeof(int), MAX_NAMELEN);
	off += MAX_NAMELEN / sizeof(int);

	//	fprintf(stderr, "P[%d].name: %s\n", i, P[i].name);
    }
    if (sizeof(int) * off != TAG_SIZE + POFF2 * (Nclient + Nenemy)) {
	fprintf(stderr,
		"unpack_playerinfo2 off:%d  TAG_SIZE+POFF2*(Nclient+Nenemy):%d\n"
		"something wrong. abort...\n",
		sizeof(int) * off, TAG_SIZE + POFF2 * (Nclient + Nenemy));
    }
    // printf("cnt: %d x:%+03d  y:%+03d  dir:%3d\n",
    // cnt++, Me->x, Me->y, Me->dir);
}

static void
unpack_stageinfo(void)
{
    int off;
    off = 2; // offset for tag
    Stage = ntohl(((int *)(RecvBuf))[off++]);
    Nflag = ntohl(((int *)(RecvBuf))[off++]);
    NflagLeft = ntohl(((int *)(RecvBuf))[off++]);
    Nenemy = ntohl(((int *)(RecvBuf))[off++]);

    //    fprintf(stderr, "Stage:%d  Nflag:%d  Nenemy:%d\n",
    //	    Stage, Nflag, Nenemy);
}

static void
unpack_map(void)
{
    int i, j;
    for (i = 0; i < MAP_HEIGHT; i++) {
	for (j = 0; j < MAP_WIDTH; j++) {
	    Map[i][j] = RecvBuf[TAG_SIZE + MAP_WIDTH * i + j];
	}
    }
    // print_map();
}

static void
unpack_smoke(void)
{
    int i, off;
    static int cnt = 0;
    off = 2; // offset for tag
    for (i = 0; i < MAX_SMOKE; i++) {
	Smoke[i][0] = ntohl(((int *)(RecvBuf))[off++]);
	Smoke[i][1] = ntohl(((int *)(RecvBuf))[off++]);
	Smoke[i][2] = ntohl(((int *)(RecvBuf))[off++]);
    }
}

static void
unpack_gametitle(void)
{
    GsCnt = ntohl(((int *)(RecvBuf))[2]);
    Gs = GAME_TITLE;
    //    fprintf(stderr, "GAME_TITLE %d\n", GsCnt);
}

static void
unpack_gameready(void)
{
    GsCnt = ntohl(((int *)(RecvBuf))[2]);
    if (Me->status == P_NOTLOGGEDIN) {
	Gs = GAME_RETIRE;
    }
    else {
	Gs = GAME_READY;
    }
}

static void
unpack_gamestart(void)
{
    Stage = ntohl(((int *)(RecvBuf))[2]);
    GsCnt = 0;
    if (Me->status == P_NOTLOGGEDIN) {
	Gs = GAME_RETIRE;
    }
    else {
	Gs = GAME_ONPLAY;
    }
}

static void
unpack_gamefinish(void)
{
    if (Me->status != P_NOTLOGGEDIN) {
	Gs = GAME_FINISH;
    }
    GsCnt = 0;
}

static void
unpack_gameover(void)
{
    if (Me->status != P_NOTLOGGEDIN) {
	Gs = GAME_OVER;
    }
    GsCnt = 0;
}

static void
unpack_gamequit(void)
{
    die(0, 1, &Soc);
}

static void
unpack_gamepause(void)
{
    if (Gs == GAME_PAUSE) {
	Gs = GAME_ONPLAY;
    }
    else if (Gs == GAME_ONPLAY) {
	Gs = GAME_PAUSE;
    }
}

static void
unpack_gameresult(void)
{
    Gs = GAME_RESULT;
    GsCnt = 0;
    Winner = ntohl(((int *)(RecvBuf))[2]);
}

static void
unpack_gamefinalresult(void)
{
    Gs = GAME_FINALRESULT;
    GsCnt = 0;
}

static void
unpack_gameretire(void)
{
    Gs = GAME_RETIRE;
    GsCnt = 0;
}

static void
unpack_highscore(void)
{
    Highscore = ntohl(((int *)(RecvBuf))[2]);
}

static void
game_title(void)
{
    if (GsCnt == -1) { // initialize title
	glutDisplayFunc(myDisplayTitle);
    }
    if (GsCnt == -3) { // about to start
	glutDisplayFunc(myDisplayOnPlay);
	Nflag = 0;
    }
}

static void
game_onplay(void)
{
    int i;

    GsCnt++;

    // decay existing smokes
    for (i = 0; i < MAX_SMOKE; i++) {
	if (Smoke[i][2] > 10) Smoke[i][2]--;
    }
}

static void
game_result(void)
{
    if (GsCnt == 0) { // initialize title
	glutDisplayFunc(myDisplayResult);
    }
}

static void
game_pause(void)
{
    // nothing to do
}

static void
game_finalresult(void)
{
    if (GsCnt == 0) { // initialize title
	glutDisplayFunc(myDisplayFinalResult);
    }
}

static void
game_ready(void)
{
    if (GsCnt == 0) {
	glutDisplayFunc(myDisplayOnPlay);
	GsCnt = 3;
	Nflag = 0;
    }
}

static void
game_retire(void)
{
    if (GsCnt == 0) {
	glutDisplayFunc(myDisplayRetire);
    }
    GsCnt++;
}

static void
game_finish(void)
{
    GsCnt++;
}

static void
game_over(void)
{
    GsCnt++;
}

static void
myTimer(int id)
{
    recv_from_server();
    (game_func[Gs])();

    if (Gs < 0 || Gs >= NGAME) {
	fprintf(stderr, "unknown game status: %d\n", Gs);
    }
    glutPostRedisplay();
    glutTimerFunc(10, myTimer, 0);
}

/*
 * OpenGL event handlers
 */

static void
myDisplayTitle(void)
{
    // preprocess
    glClear(GL_COLOR_BUFFER_BIT);

    glViewport(0, 0, WinWidth, WinHeight);

    putTitle();

    // postprocess
    glutSwapBuffers();
}

static void
myDisplayResult(void)
{
    // preprocess
    glClear(GL_COLOR_BUFFER_BIT);

    glViewport(0, 0, WinWidth, WinHeight);

    putResult();

    // postprocess
    glutSwapBuffers();
}

static void
myDisplayFinalResult(void)
{
    // preprocess
    glClear(GL_COLOR_BUFFER_BIT);

    glViewport(0, 0, WinWidth, WinHeight);

    putFinalResult();

    // postprocess
    glutSwapBuffers();
}

static void
myDisplayOnPlay(void)
{
    int i;
    double x, y;

    // preprocess
    glClear(GL_COLOR_BUFFER_BIT);

    glViewport(0, 0, WinWidth, WinHeight);
    putInfo();
    putScore();

    // console (right hand side of the screen)
    putFuel();
    putRader();
    for (i = 0; i < Me->nlife; i++) {
	putMyCarAt(0.6 + 0.08 * i, -0.8);
    }
    for (i = 0; i < Nflag; i++) {
	x = 0.6 + 0.08 * (i % 5);
	y = 0.6 - 0.08 * (i / 5);
	if (i < Me->nflag) {
	    putFlagAt(x, y, 1.0, 0);
	}
	else if (i < Nflag - NflagLeft) {
	    putFlagAt(x, y, 1.0, 1);
	}
	else {
	    putFlagAt(x, y, 0.5, 0);
	}
    }

    // play field (left hand side of the screen)
    display_play_field();

    // postprocess
    glutSwapBuffers();
}

static void
myDisplayRetire(void)
{
    int i;
    double x, y;

    // preprocess
    glClear(GL_COLOR_BUFFER_BIT);

    glViewport(0, 0, WinWidth, WinHeight);

    // console (right hand side of the screen)
    putRader();

    for (i = 0; i < Nflag; i++) {
	x = 0.6 + 0.08 * (i % 5);
	y = 0.6 - 0.08 * (i / 5);
	if (i < Me->nflag) {
	    putFlagAt(x, y, 1.0, 0);
	}
	else if (i < Nflag - NflagLeft) {
	    putFlagAt(x, y, 1.0, 1);
	}
	else {
	    putFlagAt(x, y, 0.5, 0);
	}
    }

    // play field (left hand side of the screen)
    display_play_field();

    // postprocess
    glutSwapBuffers();
}

static void
display_play_field(void)
{
    int i, j, i0, j0, i1, j1;

    glViewport(10 * WinWidth / WIN_WIDTH, 10 * WinHeight / WIN_HEIGHT,
	       320 * WinWidth / WIN_WIDTH, 320 * WinHeight / WIN_HEIGHT);
    putBackground();

    // set field of view of field map.
    i0 = PF2M_SCALE(Me->y) - MAP_FOV;
    if (i0 < 0) i0 = 0;
    j0 = PF2M_SCALE(Me->x) - MAP_FOV;
    if (j0 < 0) j0 = 0;
    i1 = PF2M_SCALE(Me->y) + MAP_FOV + 1;
    if (i1 > MAP_HEIGHT) i1 = MAP_HEIGHT;
    j1 = PF2M_SCALE(Me->x) + MAP_FOV + 1;
    if (j1 > MAP_WIDTH) j1 = MAP_WIDTH;

    // read objects from the map and put them on the play field.
    for (i = i0; i < i1; i++) {
	for (j = j0; j < j1; j++) {
	    switch(Map[i][j]) {
	      case MAP_NONE:
		putRoad(M2GL(j, i));
		break;
	      case MAP_WALL:
		putWall(j, i);
		break;
	      case MAP_FLAG:
		putFlag(M2GL(j, i));
		break;
	      case MAP_LUCKY:
		putLuckyFlag(M2GL(j, i));
		break;
	      case MAP_SPECIAL:
		putSpecialFlag(M2GL(j, i));
		break;
	      case MAP_ROCK:
		putRock(M2GL(j, i));
		break;
	      default:
		break;
	    }
	}
    }
    for (i = i0; i < i1; i++) {
	for (j = j0; j < j1; j++) {
	    switch(Map[i][j]) {
	      case MAP_WALL:
		putWallBorder(j, i);
		break;
	      default:
		break;
	    }
	}
    }

    // put cars
    for (i = 0; i < Nclient + Nenemy; i++) {
	Player *p = P + i;
	if (p == Me) continue; // skip my car
	if (p->status == P_RETIRE ||
	    p->status == P_NOTLOGGEDIN) continue; // skip retired player
	putCar(p);
    }
    if (Me->status != P_RETIRE && Me->status != P_NOTLOGGEDIN) {
	putMyCar();
    }

    // put smokes
    for (i = 0; i < MAX_SMOKE; i++) {
	if (Smoke[i][2] == 0) continue;
	putSmoke(PF2GL(Smoke[i][0], Smoke[i][1]), Smoke[i][2]);
    }

    // put messages (eg READY, FUEL ALMOST EMPTY) over the play field
    putMessages();
}

static void
myKeyboardUp(unsigned char key, int x, int y)
{
    switch (key) {
      case ' ':
	Me->key &= ~MASK_SPACEKEY;
	break;
      case 'q':
	Me->key &= ~MASK_QKEY;
	break;
      case 's':
	Me->key &= ~MASK_SKEY;
	break;
      default:
	break;
    }
    pack_keyinput(Me->key);
}

static void
myKeyboard(unsigned char key, int x, int y)
{
    switch (key) {
      case 'q':
	Me->key |= MASK_QKEY;
	break;
      case ' ':
	Me->key |= MASK_SPACEKEY;
	break;
      case 's':
	Me->key |= MASK_SKEY;
	break;
      default:
	break;
    }
    pack_keyinput(Me->key);
}

static void
mySpecialUp(int key, int x, int y)
{
    switch (key) {
      case GLUT_KEY_LEFT:
        Me->key &= ~MASK_LEFTKEY;
	break;
      case GLUT_KEY_RIGHT:
        Me->key &= ~MASK_RIGHTKEY;
	break;
      case GLUT_KEY_UP:
        Me->key &= ~MASK_UPKEY;
	break;
      case GLUT_KEY_DOWN:
        Me->key &= ~MASK_DOWNKEY;
	break;
      default:
	break;
    }
    pack_keyinput(Me->key);
}

static void
mySpecial(int key, int x, int y)
{
    if (Gs == GAME_PAUSE) return;

    switch (key) {
      case GLUT_KEY_LEFT:
        Me->key |= MASK_LEFTKEY;
	break;
      case GLUT_KEY_RIGHT:
        Me->key |= MASK_RIGHTKEY;
	break;
      case GLUT_KEY_UP:
        Me->key |= MASK_UPKEY;
	break;
      case GLUT_KEY_DOWN:
        Me->key |= MASK_DOWNKEY;
	break;
      default:
	break;
    }
    pack_keyinput(Me->key);
}

static void
myReshape(int w, int h)
{
    double ratio = WIN_WIDTH/(double)WIN_HEIGHT;

    if (w > (int)(h * ratio)) {
	w = (int)(h * ratio);
    }
    else if (w < (int)(h * ratio)) {
	h = (int)(w / ratio);
    }
    WinWidth = w;
    WinHeight = h;
    glutReshapeWindow(w, h);
    glViewport(0, 0, WinWidth, WinHeight);
}

/*
 * character drawers
 */

static void
putSmoke0(double x, double y)
{
    glPushMatrix();
    glTranslated(x, y, 0.0);

    glBegin(GL_POLYGON);
    glVertex2d(-0.5, -1.0);
    glVertex2d(-1.0, -0.5);

    glVertex2d(-1.0, +0.5);
    glVertex2d(-0.5, +1.0);

    glVertex2d(+0.5, +1.0);
    glVertex2d(+1.0, +0.5);

    glVertex2d(+1.0, -0.5);
    glVertex2d(+0.5, -1.0);
    glEnd();

    glPopMatrix();
}

static void
putSmoke(double x, double y, int lifetime)
{
    double r;

    r = 1.2 * lifetime / 100.0;
    glPushMatrix();
    glTranslated(x, y, 0.0);
    glScaled(M2GL_SCALE(r), M2GL_SCALE(r), 0.0);

    glEnable(GL_BLEND);
    // glBlendFunc(GL_ONE , GL_ONE);
    glBlendFunc(GL_SRC_ALPHA , GL_ONE);

    glColor4d(0.8, 0.8, 0.8, 0.2);

    putSmoke0(-0.3, -0.3);
    putSmoke0(-0.7, +0.5);
    putSmoke0(+0.3, +0.6);
    putSmoke0(+0.7, -0.5);

    glDisable(GL_BLEND);

    glPopMatrix();
}

static void
putRoad(double x, double y)
{
    glPushMatrix();
    glTranslated(x, y, 0.0);
    glScaled(M2GL_SCALE(1), M2GL_SCALE(1), 0.0);

    glColor3d(0.8, 0.5, 0.5);
    glBegin(GL_POLYGON);
    glVertex2d(-1.0, -1.0);
    glVertex2d(-1.0, +1.0);
    glVertex2d(+1.0, +1.0);
    glVertex2d(+1.0, -1.0);
    glEnd();

    glPopMatrix();
}

static void
putWall0(void)
{
    glBegin(GL_POLYGON);
    glVertex2d(-1.0, -1.0);
    glVertex2d(-1.0, +1.0);
    glVertex2d(+1.0, +1.0);
    glVertex2d(+1.0, -1.0);
    glEnd();
}

static void
putWallBorder0(double angle)
{
    glPushMatrix();
    glRotatef(angle, 0.0, 0.0, 1.0);
    glColor3d(0.3, 0.3, 0.8);
    glLineWidth(2.0);
    glBegin(GL_LINES);
    glVertex2d(-1.0, +1.0);
    glVertex2d(+1.0, +1.0);
    glEnd();
    glPopMatrix();
}

static void
putWall(int mx, int my)
{
    glPushMatrix();
    glTranslated(M2GL(mx, my), 0.0);
    glScaled(M2GL_SCALE(1), M2GL_SCALE(1), 0.0);

    glColor3d(0.8, 0.2, 0.8);

    putWall0();

    glPopMatrix();
}

static void
putWallBorder(int mx, int my)
{
    glPushMatrix();
    glTranslated(M2GL(mx, my), 0.0);
    glScaled(M2GL_SCALE(1), M2GL_SCALE(1), 0.0);

    if (my > 0 && Map[my-1][mx] != MAP_WALL) {
	putWallBorder0(0);
    }
    if (my < MAP_HEIGHT-1 && Map[my+1][mx] != MAP_WALL) {
	putWallBorder0(180);
    }
    if (mx > 0 && Map[my][mx-1] != MAP_WALL) {
	putWallBorder0(90);
    }
    if (mx < MAP_WIDTH-1 && Map[my][mx+1] != MAP_WALL) {
	putWallBorder0(-90);
    }

    glPopMatrix();
}

/*
 * b: brightness
 * wc: with check mark
 */

static void
putFlagAt(double x, double y, double b, int wc)
{
    double r = 0.45;

    glPushMatrix();
    glTranslated(x, y, 0.0);
    glScaled(M2GL_SCALE(r), M2GL_SCALE(r), 0.0);

    glColor3d(1.0 * b, 1.0 * b, 0.0 * b);

    putFlag0();

    if (wc) {
	glColor3d(1.0, 0.0, 0.0);
	glLineWidth(2.0);
	glBegin(GL_LINES);
	glVertex2d(-0.9, -0.9);
	glVertex2d(+0.9, +0.9);
	glVertex2d(-0.9, +0.9);
	glVertex2d(+0.9, -0.9);
	glEnd();
    }

    glPopMatrix();
}

static void
putFlag(double x, double y)
{
    putRoad(x, y);

    glPushMatrix();
    glTranslated(x, y, 0.0);
    glScaled(M2GL_SCALE(1), M2GL_SCALE(1), 0.0);

    glColor3d(1.0, 1.0, 0.0);

    putFlag0();

    glPopMatrix();
}

static void
putLuckyFlag(double x, double y)
{
    putRoad(x, y);

    glPushMatrix();
    glTranslated(x, y, 0.0);
    glScaled(M2GL_SCALE(1), M2GL_SCALE(1), 0.0);

    glColor3d(1.0, 1.0, 0.0);
    putFlag0();
    glColor3d(1.0, 0.0, 0.0);
    putString("L", -1.8, -0.5, 1);

    glPopMatrix();
}

static void
putSpecialFlag(double x, double y)
{
    putRoad(x, y);

    glPushMatrix();
    glTranslated(x, y, 0.0);
    glScaled(M2GL_SCALE(1), M2GL_SCALE(1), 0.0);

    glColor3d(1.0, 1.0, 0.0);
    putFlag0();
    glColor3d(1.0, 0.0, 0.0);
    putString("S", -1.8, -0.5, 1);

    glPopMatrix();
}

static void
putFlag0(void)
{
    glBegin(GL_POLYGON);
    glVertex2d(-0.6, -0.8);
    glVertex2d(-0.6, +0.8);
    glVertex2d(-0.4, +0.8);
    glVertex2d(-0.4, -0.8);
    glEnd();

    glBegin(GL_POLYGON);
    glVertex2d(-0.4, +0.8);
    glVertex2d(+0.6, +0.4);
    glVertex2d(-0.4, +0.0);
    glEnd();

    glBegin(GL_POLYGON);
    glVertex2d(-0.7, -0.8);
    glVertex2d(-0.7, -0.6);
    glVertex2d(-0.2, -0.6);
    glVertex2d(-0.2, -0.8);
    glEnd();
}

static void
putRock(double x, double y)
{
    putRoad(x, y);

    glPushMatrix();
    glTranslated(x, y, 0.0);
    glScaled(M2GL_SCALE(1), M2GL_SCALE(1), 0.0);

    glColor3d(0.25, 0.15, 0.1);

    glBegin(GL_POLYGON);
    glVertex2d(-0.9, -0.9);
    glVertex2d(-0.5, +0.8);
    glVertex2d(+0.7, -0.9);
    glEnd();

    glBegin(GL_POLYGON);
    glVertex2d(-0.7, -0.9);
    glVertex2d(+0.0, +0.6);
    glVertex2d(+0.8, -0.9);
    glEnd();

    glBegin(GL_POLYGON);
    glVertex2d(-0.5, -0.9);
    glVertex2d(+0.6, +0.7);
    glVertex2d(+0.9, -0.9);
    glEnd();

    glPopMatrix();
}

static void
putCrash(Player *p)
{
    glPushMatrix();
    glTranslated(PF2GL(p->x, p->y), 0.0);
    glScaled(M2GL_SCALE(1), M2GL_SCALE(1), 0.0);
    glRotatef((double)p->dir, 0.0, 0.0, 1.0);

    glScaled(2, 4, 0.0);
    glColor3d(0.8, 0.0, 0.0);
    glBegin(GL_POLYGON);
    glVertex2d(-0.40, +0.00);
    glVertex2d(-0.72, +0.20);
    glVertex2d(-0.32, +0.12);
    glVertex2d(-0.36, +0.52);
    glVertex2d(+0.00, +0.12);
    glVertex2d(+0.36, +0.52);
    glVertex2d(+0.32, +0.12);
    glVertex2d(+0.72, +0.20);
    glVertex2d(+0.40, +0.00);
    glEnd();
    glPopMatrix();
}

static void
putCar(Player *p)
{
    int isenemy = 1;
    if (p2id(p) < Nclient) isenemy = 0;

    glPushMatrix();
    glTranslated(PF2GL(p->x, p->y), 0.0);
    glScaled(M2GL_SCALE(1), M2GL_SCALE(1), 0.0);
    glRotatef((double)p->dir, 0.0, 0.0, 1.0);

    if (isenemy) {
	glColor3d(0.8, 0.8, 0.8);
    }
    else {
	glColor3d(1.0, 0.0, 0.0);
    }
    glBegin(GL_POLYGON);
    glVertex2d(-0.7, -0.7);
    glVertex2d(-0.3, +0.7);
    glVertex2d(+0.3, +0.7);
    glVertex2d(+0.7, -0.7);
    glEnd();

    glColor3d(0.0, 0.0, 0.0);
    glBegin(GL_POLYGON);
    glVertex2d(-0.7, -0.8);
    glVertex2d(-0.7, -0.1);
    glVertex2d(-0.3, -0.1);
    glVertex2d(-0.3, -0.8);
    glEnd();

    glColor3d(0.0, 0.0, 0.0);
    glBegin(GL_POLYGON);
    glVertex2d(+0.7, -0.8);
    glVertex2d(+0.7, -0.1);
    glVertex2d(+0.3, -0.1);
    glVertex2d(+0.3, -0.8);
    glEnd();

    glColor3d(0.0, 0.0, 0.0);
    glBegin(GL_POLYGON);
    glVertex2d(-0.6, +0.2);
    glVertex2d(-0.6, +0.7);
    glVertex2d(-0.3, +0.7);
    glVertex2d(-0.3, +0.2);
    glEnd();

    glColor3d(0.0, 0.0, 0.0);
    glBegin(GL_POLYGON);
    glVertex2d(+0.6, +0.2);
    glVertex2d(+0.6, +0.7);
    glVertex2d(+0.3, +0.7);
    glVertex2d(+0.3, +0.2);
    glEnd();

    glPopMatrix();

    if (p->status == P_CRASH) {
	putCrash(p);
	return;
    }
}

static void
putMyCarAt(double x, double y)
{
    double r = 0.5;

    glPushMatrix();
    glTranslated(x, y, 0.0);
    glScaled(M2GL_SCALE(r), M2GL_SCALE(r), 0.0);

    glColor3d(0.0, 0.0, 0.5);
    glBegin(GL_POLYGON);
    glVertex2d(-1.0, -1.0);
    glVertex2d(-1.0, +1.0);
    glVertex2d(+1.0, +1.0);
    glVertex2d(+1.0, -1.0);
    glEnd();

    putMyCar0();

    glPopMatrix();
}

static void
putMyCar(void)
{
    //    fprintf(stderr, "Me->dir %d\n", Me->dir);
    glPushMatrix();
    glScaled(M2GL_SCALE(1), M2GL_SCALE(1), 0.0);
    glRotatef((double)Me->dir, 0.0, 0.0, 1.0);

    putMyCar0();

    glPopMatrix();

    if (Me->status == P_CRASH) {
	putCrash(Me);
	return;
    }
}

static void
putMyCar0(void)
{
    glColor3d(0.0, 0.0, 1.0);
    glBegin(GL_POLYGON);
    glVertex2d(-0.7, -0.7);
    glVertex2d(-0.2, +0.7);
    glVertex2d(+0.2, +0.7);
    glVertex2d(+0.7, -0.7);
    glEnd();

    glColor3d(0.0, 0.0, 0.0);
    glBegin(GL_POLYGON);
    glVertex2d(-0.7, -0.8);
    glVertex2d(-0.7, -0.1);
    glVertex2d(-0.3, -0.1);
    glVertex2d(-0.3, -0.8);
    glEnd();

    glColor3d(0.0, 0.0, 0.0);
    glBegin(GL_POLYGON);
    glVertex2d(+0.7, -0.8);
    glVertex2d(+0.7, -0.1);
    glVertex2d(+0.3, -0.1);
    glVertex2d(+0.3, -0.8);
    glEnd();

    glColor3d(0.0, 0.0, 0.0);
    glBegin(GL_POLYGON);
    glVertex2d(-0.6, +0.2);
    glVertex2d(-0.6, +0.7);
    glVertex2d(-0.3, +0.7);
    glVertex2d(-0.3, +0.2);
    glEnd();

    glColor3d(0.0, 0.0, 0.0);
    glBegin(GL_POLYGON);
    glVertex2d(+0.6, +0.2);
    glVertex2d(+0.6, +0.7);
    glVertex2d(+0.3, +0.7);
    glVertex2d(+0.3, +0.2);
    glEnd();
}

static void
putBackground(void)
{
    // glColor3d(0.5, 0.3, 0.2);
    //    glColor3d(0.4, 0.5, 0.99);
    glColor3d(0.3, 0.5, 0.2);
    glBegin(GL_POLYGON);
    glVertex2d(-1.0, -1.0);
    glVertex2d(-1.0, +1.0);
    glVertex2d(+1.0, +1.0);
    glVertex2d(+1.0, -1.0);
    glEnd();
}

static void
putInfo(void)
{
    char buf[128];

#if 0 // just for debugging
    glColor3d(1.0, 1.0, 0.0);
    sprintf(buf, "x:%5d  y:%5d  dir:%+4d  flag:%d  life:%d  fuel:%d",
	    Me->x, Me->y, Me->dir, Me->nflag, Me->nlife, Me->fuel);
    putString(buf, -0.9, 0.90, 2);
    sprintf(buf, "recv hit:%5d try:%5d (%3d%%)",
	    Statistics.hit, 
	    Statistics.try, 
	    100 * Statistics.hit / Statistics.try);
    putString(buf, -0.9, 0.8, 2);
#endif
    glColor3d(1.0, 1.0, 0.0);
    sprintf(buf, "ROUND %d", Stage);
    putString(buf, 0.55, -0.93, 2);
}

static void
putScore(void)
{
    char buf[128];

    glColor3d(1.0, 1.0, 0.0);
    sprintf(buf, "SCORE: %08d    HISCORE: %08d",
	    Me->score, Highscore);
    putString(buf, -0.9, 0.93, 2);
}

static void
putRader(void)
{
    int i, j;
    static blinkcnt = 0;

    double w, h, rx, ry;
    w = 0.2;
    h = w * MAP_HEIGHT / MAP_WIDTH;
    rx = w / (PF_WIDTH * 0.5);
    ry = -h / (PF_HEIGHT * 0.5);

    glPushMatrix();
    glTranslated(0.75, -0.3, 0.0);
    glScaled(rx, ry, 0.0);
    glTranslated(-PF_WIDTH * 0.5, -PF_HEIGHT * 0.5, 0.0);

    glColor3d(0.0, 0.0, 0.5);
    glBegin(GL_POLYGON);
    glVertex2d(+0.0, +0.0);
    glVertex2d(+0.0, PF_HEIGHT);
    glVertex2d(PF_WIDTH, PF_HEIGHT);
    glVertex2d(PF_WIDTH, +0.0);
    glEnd();

    if (Me->status != P_RETIRE && Me->status != P_NOTLOGGEDIN) {
	glColor3d(1.0, 0.0, 0.0);
	glPointSize(5.0);
	glBegin(GL_POINTS);
	glVertex2d(Me->x, Me->y);
	glEnd();
    }

    // put car(s)
    for (i = 0; i < Nclient + Nenemy; i++) {
	Player *p = P + i;
	if (p == Me) continue; // skip my car
	if (p->status == P_RETIRE ||
	    p->status == P_NOTLOGGEDIN) continue; // skip retired player
	glColor3d(1.0, 1.0, 1.0);
	glPointSize(5.0);
	glBegin(GL_POINTS);
	glVertex2d(p->x, p->y);
	glEnd();
    }

    // put object(s)
    blinkcnt++;
    for (i = 0; i < MAP_HEIGHT; i++) {
	for (j = 0; j < MAP_WIDTH; j++) {
	    switch (Map[i][j]) {
	      case MAP_FLAG:
		glColor3d(1.0, 1.0, 0.0);
		glPointSize(5.0);
		glBegin(GL_POINTS);
		glVertex2d(M2PF(j, i));
		glEnd();
		break;
	      case MAP_SPECIAL:
	      case MAP_LUCKY:
		if ((blinkcnt / 10) % 3 != 0) {
		    glColor3d(1.0, 1.0, 0.0);
		    glPointSize(5.0);
		    glBegin(GL_POINTS);
		    glVertex2d(M2PF(j, i));
		    glEnd();
		}
		break;
	      case MAP_ROCK:
		glColor3d(0.25, 0.15, 0.1);
		glPointSize(5.0);
		glBegin(GL_POINTS);
		glVertex2d(M2PF(j, i));
		glEnd();
		break;
	      default:
		break;
	    }
	}
    }

    glPopMatrix();
}

static void
putFuel(void)
{
    char buf[128];
    double w, h, r;

    w = 0.4;
    h = 0.07;
    r = Me->fuel / (double)MAX_FUEL;

    glPushMatrix();

    glTranslated(0.55, 0.15, 0.0);
    glScaled(w, h, 0.0);

    glColor3d(0.8, 0.8, 0.8);
    glBegin(GL_POLYGON);
    glVertex2d(-0.01, 0.0);
    glVertex2d(-0.01, 1.0);
    glVertex2d(+1.01, 1.0);
    glVertex2d(+1.01, 0.0);
    glEnd();

    glColor3d(0.0, 0.0, 0.5);
    glBegin(GL_POLYGON);
    glVertex2d(0.0, 0.1);
    glVertex2d(0.0, 0.9);
    glVertex2d(1.0, 0.9);
    glVertex2d(1.0, 0.1);
    glEnd();

    glColor3d(0.0, 1.0, 0.0);
    glBegin(GL_POLYGON);
    glVertex2d(1.0, 0.1);
    glVertex2d(1.0, 0.9);
    glVertex2d(1.0 - r, 0.9);
    glVertex2d(1.0 - r, 0.1);
    glEnd();

    glPopMatrix();

    glColor3d(0.0, 1.0, 0.0);
    sprintf(buf, "fuel");
    putString(buf, 0.69, 0.25, 2);
}

static void
putMessages(void)
{
    char buf[128];
    double gly = 0.2;
    static blinkcnt = 0;

    switch (Gs) {
      case GAME_ONPLAY:
	if (Me->fuel < ALMOST_EMPTY_FUEL &&
	    Me->status != P_FUELEMPTY) {
	    static blinkcnt = 0;
	    blinkcnt++;
	    if ((blinkcnt / 10) % 3 != 0) {
		glColor3d(1.0, 0.7, 0.0);
		putString("FUEL ALMOST EMPTY", -0.6, gly, 1);
	    }
	}
	switch (Me->status) {
	  case P_NORMAL:
	    if (GsCnt < 200) {
		glColor3d(1.0, 0.0, 0.0);
		putString("GO!", -0.1, gly, 1);
	    }
	    break;
	  case P_FUELEMPTY:
	    glColor3d(1.0, 0.0, 0.0);
	    putString("FUEL EMPTY", -0.40, gly, 1);
	    break;
	  default:
	    break;
	}
	break;
      case GAME_READY:
	    glColor3d(0.0, 1.0, 0.0);
	    sprintf(buf, "ROUND %d", Stage);
	    putString(buf, -0.25, gly+0.15, 1);
	    sprintf(buf, "READY  %d", GsCnt);
	    putString(buf, -0.25, gly, 1);
	break;
      case GAME_RETIRE:
	glColor3d(0.0, 1.0, 0.0);
	blinkcnt++;
	if ((blinkcnt / 20) % 3 != 0) {
	    sprintf(buf, "LIVE");
	    putString(buf, +0.6, +0.8, 1);
	}
	break;
      case GAME_FINISH:
	glColor3d(0.0, 1.0, 0.0);
	putString("STAGE CLEARED", -0.5, gly, 1);
	break;
      case GAME_OVER:
	glColor3d(1.0, 0.0, 0.0);
	putString("GAME OVER", -0.35, gly, 1);
	break;
      case GAME_PAUSE:
	glColor3d(0.0, 1.0, 0.0);
	putString("PAUSE", -0.20, gly, 1);
	break;
      default:
	break;
    }
}

static void
putTitle(void)
{
    int i;
    double x, y, y0;
    char buf[128];

    y0 = +0.70;
    glColor3d(0.7, 0.7, 1.0);
    putString("new", -0.7, y0+0.03, 1);
    putString("Rally-Z", -0.50, y0, 0);
    glColor3d(1.0, 0.7, 0.0);
    putString("online", -0.12, y0-0.01, 1);
    glColor3d(0.0, 0.7, 0.0);
    sprintf(buf, "ver. %08x", RALLYZ_VERSION);
    putString(buf, +0.6, +0.95, 2);

    glColor3d(0.0, 1.0, 0.0);
    putString("Car Control:", -0.5, -0.83, 2);
    if (Me->ctrlstyle == P_CLASSIC) {
	glColor3d(0.0, 1.0, 0.0);
	putString("CLASSIC", -0.1, -0.83, 2);
	glColor3d(0.0, 0.5, 0.0);
	putString("MODERN", +0.3, -0.83, 2);
    }
    else {
	glColor3d(0.0, 0.5, 0.0);
	putString("CLASSIC", -0.1, -0.83, 2);
	glColor3d(0.0, 1.0, 0.0);
	putString("MODERN", +0.3, -0.83, 2);
    }

    if (Me->status == P_NOTLOGGEDIN) {
	static blinkcnt = 0;
	blinkcnt++;
	if ((blinkcnt / 20) % 3 != 0) {
	    glColor3d(1.0, 1.0, 0.0);
	    sprintf(buf, "Welcome %s. Press SPACE key to check in.", Me->name);
	    putString(buf, -0.7, -0.7, 2);
	}
    }
    else {
	glColor3d(1.0, 1.0, 0.0);
	putString("Waiting for Other Guys...", -0.7, -0.7, 2);
    }
    if (GsCnt > 0) {
	static blinkcnt = 0;
	blinkcnt++;
	if ((blinkcnt / 20) % 3 != 0) {
	    glColor3d(1.0, 0.7, 0.0);
	    sprintf(buf, "Game Starts in %d sec", GsCnt);
	    putString(buf, -0.3, -0.95, 2);
	}
    }

    x = -0.7;
    y = y0 = 0.3;

    glColor3d(0.0, 0.7, 0.0);
    glLineWidth(1.0);
    glBegin(GL_LINES);
    glVertex2d(x-0.03, y0+0.18);
    glVertex2d(0.7, y0+0.18);
    glEnd();

    glColor3d(0.0, 1.0, 0.0);
    putString("player", x, y0+0.11, 2);
    putString("wins", x+0.5, y0+0.11, 2);
    putString("score", x+0.75, y0+0.11, 2);
    //    putString("check in", x+1.15, y0+0.11, 2);

    glColor3d(0.0, 0.7, 0.0);
    glLineWidth(1.0);
    glBegin(GL_LINES);
    glVertex2d(x-0.03, y0+0.07);
    glVertex2d(0.7, y0+0.07);
    glEnd();

    glColor3d(0.0, 1.0, 0.0);
    for (i = 0; i < Nclient; i++, y -= 0.1) {
	Player *p = P + i;

	if (p == Me) {
	    glColor3d(1.0, 1.0, 0.0);
	}
	else {
	    glColor3d(0.0, 1.0, 0.0);
	}
	sprintf(buf, "%s", p->name);
	putString(buf, x, y, 2);
	switch (p->nwin) {
	  case 0:
	    sprintf(buf, "  -  ", p->nwin);
	    break;
	  case 1:
	    sprintf(buf, "% 2d", p->nwin);
	    break;
	  default:
	    sprintf(buf, "% 2d", p->nwin);
	    break;
	}
	putString(buf, x+0.5, y, 2);
	sprintf(buf, "%08d", p->score);
	putString(buf, x+0.75, y, 2);
	if (p->status == P_NOTLOGGEDIN) {
	    glColor3d(0.5, 0.35, 0.0);
	    putString("checkin", x+1.15, y, 2);
	}
	else {
	    glColor3d(1.0, 0.7, 0.0);
	    putString("checkin", x+1.15, y, 2);
	}
    }

    glColor3d(0.0, 0.7, 0.0);
    glLineWidth(1.0);
    glBegin(GL_LINES);
    glVertex2d(x-0.03, y+0.07);
    glVertex2d(0.7, y+0.07);
    glEnd();

}

static void
putResult(void)
{
    char buf[128];
    int i;
    Player *p;
    double y;

    glColor3d(0.7, 0.7, 1.0);
    sprintf(buf, "Round %d Result", Stage);
    putString(buf, -0.7, +0.5, 1);

    for (i = 0, y = 0.2; i < Nclient; i++) {
	p = P + i;
	if (p->status == P_NOTLOGGEDIN) continue;
	y -= 0.2;
	if (p == Me) {
	    glColor3d(0.0, 1.0, 0.0);
	}
	else {
	    glColor3d(0.0, 0.7, 0.0);
	}
	if (i == Winner) {
	    glColor3d(1.0, 1.0, 0.0);
	}

	sprintf(buf, "%s", p->name);
	putString(buf, -0.85, y, 2);

	sprintf(buf, "x %d", p->nflag);
	putString(buf, -0.22, y, 2);

	sprintf(buf, "%08d pts", p->score);
	putString(buf, +0.00, y, 2);

	if (p->status == P_RETIRE) {
	    sprintf(buf, "RETIRED");
	}
	else {
	    sprintf(buf, "FINISHED");
	}
	putString(buf, +0.50, y, 2);

	glColor3d(1.0, 1.0, 0.0);
	glPushMatrix();
	glTranslated(-0.28, y + 0.023, 0.0);
	glScaled(M2GL_SCALE(0.8), M2GL_SCALE(0.8), 0.0);
	putFlag0();
	glPopMatrix();
    }
}

static void
putFinalResult(void)
{
    char buf[128];
    int i;
    Player *p;
    double y;

    glColor3d(0.7, 0.7, 1.0);
    sprintf(buf, "Final Result");
    putString(buf, -0.7, +0.5, 1);

    for (i = 0, y = 0.2; i < Nclient; i++) {
	p = P + i;
	if (p->status == P_NOTLOGGEDIN) continue;
	y -= 0.2;
	if (p == Me) {
	    glColor3d(0.0, 1.0, 0.0);
	}
	else {
	    glColor3d(0.0, 0.7, 0.0);
	}

	sprintf(buf, "%s", p->name);
	putString(buf, -0.85, y, 2);

	sprintf(buf, "x %d", p->ntotalflag);
	putString(buf, -0.22, y, 2);

	sprintf(buf, "%08d pts", p->score);
	putString(buf, +0.00, y, 2);

	glColor3d(1.0, 1.0, 0.0);
	glPushMatrix();
	glTranslated(-0.28, y + 0.023, 0.0);
	glScaled(M2GL_SCALE(0.8), M2GL_SCALE(0.8), 0.0);
	putFlag0();
	glPopMatrix();
    }
}

static int
p2id(Player *p)
{
    int i;
    for (i = 0; i < Nclient + Nenemy; i++) {
	if (p == P + i) break;
    }
    return i;
}

// just for debugging purpose
static void
print_map(void)
{
    int i, j;
    char c;

    for (i = 0; i < MAP_HEIGHT; i++) {
	printf("\n", c);
	for (j = 0; j < MAP_WIDTH; j++) {
	    switch (Map[i][j]) {
	      case MAP_NONE:
		c = '.';
		break;
	      case MAP_WALL:
		c = '*';
		break;
	      case MAP_FLAG:
		c = 'f';
		break;
	      case MAP_ROCK:
		c = 'r';
		break;
	      case MAP_LUCKY:
		c = 'l';
		break;
	      case MAP_SPECIAL:
		c = 's';
		break;
	      default:
		c = '?';
		break;
	    }
	    printf("%c", c);
	}
    }
    printf("\n", c);
}
