// server-side main module

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <sys/types.h>
#ifndef WIN32
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>
#endif
#include "utils.h"
#include "commondefs.h"

#define WIN_WIDTH (300)  // window size
#define WIN_HEIGHT (100)

static int Nclient; // number of clients
static Gamestatus Gs;
static int GsCnt; // general-purpose counter
static char Map[MAP_HEIGHT][MAP_WIDTH];
static int Nflag;
static int NflagLeft;
static int Smoke[MAX_SMOKE][3]; // position & lifetime of smokes
static int Stage;
static int Nenemy;
static int Highscore = 0;
static int TimerPeriod = INIT_TIMER_PERIOD;

static SOCKET Soc[MAX_CLIENT]; // socket for each client
static char RecvBuf[MAX_DATA_SIZE];
static char SendBuf[MAX_DATA_SIZE];
static Player P[MAX_CLIENT];
static void (* game_func[NGAME])();

// data transfer routines
static void init_vars(void);
static void login_to_session(int id, SOCKET fd);
static void init_session(int n);
static void recv_from_clients(void);
static void broadcast_to_clients(int size);
static void pack_playerinfo(void);
static void pack_playerinfo2(void);
static void pack_stageinfo(void);
static void pack_smoke(void);
static void pack_highscore(int hs);
static void pack_gamepause(void);
static void pack_gametitle(int cnt);
static void pack_gameready(int cnt);
static void pack_gamestart(int stage);
static void pack_gamefinish(void);
static void pack_gameresult(int winner);
static void pack_gamefinalresult(void);
static void pack_gameretire(Player *p);
static void pack_gameover(void);
static void pack_gamequit(void);

// state machine
static void game_title(void);
static void game_ready(void);
static void game_onplay(void);
static void game_pause(void);
static void game_finish(void);
static void game_over(void);
static void game_result(void);
static void game_finalresult(void);
static void toggle_pause(void);

// routines to manage game events
static void move_player(void);
static void move_position4retired(Player *p);
static void move_position_modern(Player *p);
static void move_direction_modern(Player *p);
static void move_position_classic(Player *p);
static void move_enemy_position(Player *p);
static void check_hit(void);
static void check_fuel(void);
static void crash_car(Player *p);
static void check_finish(void);
static void check_over(void);
static int recover_from_crash(Player *p);
static void spin_car(Player *p);
static int recover_from_spin(Player *p);
static void generate_smoke(Player *p);
static void decay_smoke(void);
static void control_enemy(void);

// utilities
static void init_player(int id, unsigned int flag);
static void load_map(int stage);
static int p2id(Player *p);

// OpenGL event handlers
static void myDisplay(void);
static void myReshape(int w, int h);
static void myTimer(int id);
static void myKeyboard(unsigned char key, int x, int y);
static void kickoff_opengl(int argc, char *argv[]);

// just for debugging
static void print_map(void);

int
main(int argc, char *argv[])
{
    int num;
    int soc_waiting, maxfd;
    char buf[32];

    if (argc > 1) {
	num = atoi(argv[1]);
    }
    else {
	printf("how many clients?: ");
	fgets(buf, 16, stdin);
	sscanf(buf, "%d", &num);
    }
    Nclient = num;
    printf("OK, setup server with %d clients.\n", Nclient);

    soc_waiting = mserver_socket(PORT, num);
    if (soc_waiting == -1) {
	fprintf(stderr, "mserver_socket failed\n");
	exit(1);
    }
    maxfd = mserver_maccept(soc_waiting, num, login_to_session);
    if (maxfd == -1) {
	fprintf(stderr, "mserver_maccept failed\n");
	exit(1);
    }
    init_vars();
    init_session(num);
    kickoff_opengl(argc, argv);
    die(0, Nclient, Soc);
}

static void
init_vars(void)
{
    game_func[GAME_TITLE] = game_title;
    game_func[GAME_READY] = game_ready;
    game_func[GAME_ONPLAY] = game_onplay;
    game_func[GAME_PAUSE] = game_pause;
    game_func[GAME_FINISH] = game_finish;
    game_func[GAME_RESULT] = game_result;
    game_func[GAME_OVER] = game_over;
    game_func[GAME_FINALRESULT] = game_finalresult;
}

static void
kickoff_opengl(int argc, char *argv[])
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGB);
    glutInitWindowSize(WIN_WIDTH, WIN_HEIGHT);
    glutCreateWindow("Rally-Z server");
    glutDisplayFunc(myDisplay);
    glutTimerFunc(10, myTimer, 0);
    glutKeyboardFunc(myKeyboard);
    glutReshapeFunc(myReshape);
    glutMainLoop();
}

/*
 * invoked for each client when he/she get connected.
 *
 * id: player's id
 * fd: player's file descriptor
 */
static void
login_to_session(int id, SOCKET fd)
{
    static char *mesg = "your name: ";
    int len = strlen(mesg) + 1;

    Soc[id] = fd;
    *(int *)SendBuf = htonl(len);
    strcpy(SendBuf + sizeof(int), mesg);
    send(Soc[id], SendBuf, sizeof(int) + len, 0);
    recv(Soc[id], RecvBuf, sizeof(int), 0);
    len = ntohl(*(int *)RecvBuf);
    recv(Soc[id], P[id].name, len, 0);
    if (strlen(P[id].name) == 0) {
	sprintf(P[id].name, "Nanashi%02d", id);
    }
    if (strlen(P[id].name) > MAX_NAMELEN) {
	P[id].name[MAX_NAMELEN - 1] = 0;
    }
    *(int *)SendBuf = htonl(id);
    send(Soc[id], SendBuf, sizeof(int), 0);
    printf("[%d] %s\n", id, P[id].name);
}

/*
 * n: # of players
 */
static void
init_session(int n)
{
    int i, len;
    static char mesg[128];

    Nclient = n;

    fprintf(stderr, "Nclient: %d\n\n", Nclient);

    *(int *)SendBuf = htonl(Nclient);
    for (i = 0; i < Nclient; i++) {
	send(Soc[i], SendBuf, sizeof(int), 0);
    }

    for (i = 0; i < Nclient; i++) {
	sprintf(mesg,
		"Hello, %s. Now we start a new session.\n\n",
		P[i].name);
	len = strlen(mesg) + 1;
	*(int *)SendBuf = htonl(len);
	strcpy(SendBuf + sizeof(int), mesg);
	send(Soc[i], SendBuf, sizeof(int) + len, 0);
    }
    for (i = 0; i < MAX_CLIENT; i++) {
	init_player(i, 3);
    }
    Gs = GAME_TITLE;
    GsCnt = 0;
}

static void
recv_from_clients(void)
{
    int i, tag, size;

    for (i = 0; i < Nclient; i++) {
#if defined(WIN32)
	ioctlsocket(Soc[i], FIONREAD, &size);
	if (size < TAG_SIZE) continue;
#else
	size = recv(Soc[i], RecvBuf, TAG_SIZE, MSG_PEEK | MSG_DONTWAIT);
	if (size < 0) {
	    if (errno == EAGAIN) { // no data arrived yet.
		continue;
	    }
	    else {
		perror("recv_from_server EAGAIN");
	    }
	}
	if (size < TAG_SIZE) continue; // no data arrived yet.
#endif
	recv(Soc[i], RecvBuf, TAG_SIZE, MSG_PEEK);
	tag = ntohl(((int *)RecvBuf)[0]);
	size = ntohl(((int *)RecvBuf)[1]);
	recv(Soc[i], RecvBuf, size, 0);

	switch (tag) {
	  case TAG_KEYPRESSED:
	    P[i].key = ntohl(((int *)(RecvBuf))[2]);
	    break;
	  default:
	    fprintf(stderr, "undefined tag: %d\n", tag);
	    break;
	}
    }
}

/*
 * broadcast information of all players to all players
 */
static void
pack_playerinfo(void)
{
    int i, off, size;

    ((int *)(SendBuf))[0] = htonl(TAG_PLAYERINFO);
    off = 2;
    for (i = 0; i < Nclient + Nenemy; i++) {
	((int *)(SendBuf))[off++] = htonl(P[i].x);
	((int *)(SendBuf))[off++] = htonl(P[i].y);
	((int *)(SendBuf))[off++] = htonl(P[i].dir);
	((int *)(SendBuf))[off++] = htonl(P[i].status);
	((int *)(SendBuf))[off++] = htonl(P[i].fuel);
    }
    size = sizeof(int) * off;
    ((int *)(SendBuf))[1] = htonl(size);
    if (sizeof(int) * off != TAG_SIZE+POFF * (Nclient + Nenemy)) {
	fprintf(stderr,
		"pack_playerinfo off:%d  TAG_SIZE+POFF*(Nclient+Nenemy):%d\n"
		"something wrong. abort...\n",
		sizeof(int) * off, TAG_SIZE+POFF * (Nclient + Nenemy));
    }
    broadcast_to_clients(size);
}

static void
pack_playerinfo2(void)
{
    int i, off, size;

    ((int *)(SendBuf))[0] = htonl(TAG_PLAYERINFO2);
    off = 2;
    for (i = 0; i < Nclient + Nenemy; i++) {
	((int *)(SendBuf))[off++] = htonl(P[i].x);
	((int *)(SendBuf))[off++] = htonl(P[i].y);
	((int *)(SendBuf))[off++] = htonl(P[i].dir);
	((int *)(SendBuf))[off++] = htonl(P[i].status);
	((int *)(SendBuf))[off++] = htonl(P[i].fuel);
	((int *)(SendBuf))[off++] = htonl(P[i].nflag);
	((int *)(SendBuf))[off++] = htonl(P[i].nlife);
	((int *)(SendBuf))[off++] = htonl(P[i].nwin);
	((int *)(SendBuf))[off++] = htonl(P[i].score);
	((int *)(SendBuf))[off++] = htonl(P[i].ntotalflag);
	((int *)(SendBuf))[off++] = htonl(P[i].ctrlstyle);
	memcpy(SendBuf + off * sizeof(int), P[i].name, MAX_NAMELEN);
	off += MAX_NAMELEN / sizeof(int);
    }
    if (sizeof(int) * off  != TAG_SIZE+POFF2 * (Nclient + Nenemy)) {
	fprintf(stderr,
		"pack_playerinfo2 off:%d  TAG_SIZE+POFF2*(Nclient+Nenemy):%d\n"
		"something wrong. abort...\n",
		sizeof(int) * off, TAG_SIZE + POFF2 * (Nclient + Nenemy));
    }
    size = sizeof(int) * off;
    ((int *)(SendBuf))[1] = htonl(size);
    broadcast_to_clients(size);
}

static void
pack_stageinfo(void)
{
    int off, size;

    ((int *)(SendBuf))[0] = htonl(TAG_STAGEINFO);
    off = 2;
    ((int *)(SendBuf))[off++] = htonl(Stage);
    ((int *)(SendBuf))[off++] = htonl(Nflag);
    ((int *)(SendBuf))[off++] = htonl(NflagLeft);
    ((int *)(SendBuf))[off++] = htonl(Nenemy);
    size = sizeof(int) * off;
    ((int *)(SendBuf))[1] = htonl(size);
    broadcast_to_clients(size);
}

static void
pack_map(void)
{
    int i, j, size;

    ((int *)(SendBuf))[0] = htonl(TAG_MAP);
    size = TAG_SIZE + MAP_HEIGHT * MAP_WIDTH;
    ((int *)(SendBuf))[1] = htonl(size);
    for (i = 0; i < MAP_HEIGHT; i++) {
	for (j = 0; j < MAP_WIDTH; j++) {
	    SendBuf[TAG_SIZE + MAP_WIDTH * i + j] = Map[i][j];
	}
    }
    broadcast_to_clients(size);
}

static void
pack_smoke(void)
{
    int i, off, size;

    ((int *)(SendBuf))[0] = htonl(TAG_SMOKE);
    off = 2;
    for (i = 0; i < MAX_SMOKE; i++) {
	((int *)SendBuf)[off++] = htonl(Smoke[i][0]);
	((int *)SendBuf)[off++] = htonl(Smoke[i][1]);
	((int *)SendBuf)[off++] = htonl(Smoke[i][2]);
    }
    size = sizeof(int) * off;
    ((int *)(SendBuf))[1] = htonl(size);
    broadcast_to_clients(size);
}

static void
pack_gamequit(void)
{
    int size;
    ((int *)(SendBuf))[0] = htonl(TAG_GAMEQUIT);
    size = TAG_SIZE;
    ((int *)(SendBuf))[1] = htonl(size);
    broadcast_to_clients(size);
}

static void
pack_gamepause(void)
{
    int size;
    ((int *)(SendBuf))[0] = htonl(TAG_GAMEPAUSE);
    size = TAG_SIZE;
    ((int *)(SendBuf))[1] = htonl(size);
    broadcast_to_clients(size);
}

static void
pack_gametitle(int cnt)
{
    int size;
    ((int *)(SendBuf))[0] = htonl(TAG_GAMETITLE);
    size = TAG_SIZE + sizeof(int);
    ((int *)(SendBuf))[1] = htonl(size);
    ((int *)(SendBuf))[2] = htonl(cnt);
    broadcast_to_clients(size);
}

static void
pack_gameready(int cnt)
{
    int size;
    ((int *)(SendBuf))[0] = htonl(TAG_GAMEREADY);
    size = TAG_SIZE + sizeof(int);
    ((int *)(SendBuf))[1] = htonl(size);
    ((int *)(SendBuf))[2] = htonl(cnt);
    broadcast_to_clients(size);
}

static void
pack_gamestart(int stage)
{
    int size;
    ((int *)(SendBuf))[0] = htonl(TAG_GAMESTART);
    size = TAG_SIZE + sizeof(int);
    ((int *)(SendBuf))[1] = htonl(size);
    ((int *)(SendBuf))[2] = htonl(stage);
    broadcast_to_clients(size);
}

static void
pack_gamefinish(void)
{
    *(int *)(SendBuf) = htonl(TAG_GAMEFINISH);
    broadcast_to_clients(TAG_SIZE);
}

static void
pack_gameover(void)
{
    int size;
    ((int *)(SendBuf))[0] = htonl(TAG_GAMEOVER);
    size = TAG_SIZE;
    ((int *)(SendBuf))[1] = htonl(size);
    broadcast_to_clients(size);
}

static void
pack_gameresult(int winner)
{
    int size;
    ((int *)(SendBuf))[0] = htonl(TAG_GAMERESULT);
    size = TAG_SIZE + sizeof(int);
    ((int *)(SendBuf))[1] = htonl(size);
    ((int *)(SendBuf))[2] = htonl(winner);
    broadcast_to_clients(size);
}

static void
pack_gamefinalresult(void)
{
    int size;
    ((int *)(SendBuf))[0] = htonl(TAG_GAMEFINALRESULT);
    size = TAG_SIZE;
    ((int *)(SendBuf))[1] = htonl(size);
    broadcast_to_clients(size);
}

static void
pack_gameretire(Player *p)
{
    int size;
    ((int *)(SendBuf))[0] = htonl(TAG_GAMERETIRE);
    size = TAG_SIZE;
    ((int *)(SendBuf))[1] = htonl(size);
    send(Soc[p2id(p)], SendBuf, size, 0);
}

static void
pack_highscore(int hs)
{
    int size;
    ((int *)(SendBuf))[0] = htonl(TAG_HIGHSCORE);
    size = TAG_SIZE + sizeof(int);
    ((int *)(SendBuf))[1] = htonl(size);
    ((int *)(SendBuf))[2] = htonl(hs);
    broadcast_to_clients(size);
}

static void
broadcast_to_clients(int size)
{
    int i;

    //    fprintf(stderr, "tag:%d  size:%d\n",
    //	    ntohl(((int *)SendBuf)[0]), size);

    for (i = 0; i < Nclient; i++) {
	send(Soc[i], SendBuf, size, 0);
    }
}

/*
 * initialize a player
 *   id: player id
 *   level:
 *     0 -- recovery from crash or fuel empty
 *     1 -- reset at beginning of a new stage
 *     2 -- reset at beginning of a new game
 *     3 -- reset at beginning of this process
 */
static void
init_player(int id, unsigned int level)
{
    int i;
    int isenemy = 1;
    Player *p = P + id;

    if (id < Nclient) isenemy = 0;

    p->x = p->initx;
    p->y = p->inity;
    p->dir = 0;
    p->fuel = MAX_FUEL;
    p->vx = 0;
    p->vy = -VEL;
    p->key = 0;
    p->cnt = 0;
    for (i = 0; i < MAX_TRAIL; i++) {
	p->trail[i][0] = p->x;
	p->trail[i][1] = p->y + RES * i;
    }
    if (isenemy) {
	p->cnt = 100;
    }
    if (level > 0) { // beginning of a stage 
	p->nflag = 0;
	p->flagscore = 100;
	p->special = 1;
    }
    if (level > 1) { // beginning of a game
	if (isenemy) {
	    p->status = P_NORMAL;
	}
	else {
	    p->status = P_NOTLOGGEDIN;
	}
	p->ntotalflag = 0;
	p->nlife = INIT_NLIFE;
	p->ctrlstyle = P_CLASSIC;
    }
    if (level > 2) { // full reset
	p->nwin = 0;
    }
    pack_playerinfo2();
}

static int
p2id(Player *p)
{
    int i;
    for (i = 0; i < Nclient + Nenemy; i++) {
	if (p == P + i) break;
    }
    return i;
}

static int // 0:normal phase    1:recovery phase
recover_from_crash(Player *p)
{
    p->cnt--;
    if (p->cnt != 0) return 1; // yet in recovery phase
    if (p->nlife < 0) {
	p->status = P_RETIRE;
	pack_playerinfo2();
	pack_gameretire(p);
    }
    else {
	p->status = P_NORMAL;
	init_player(p2id(p), 0);
    }
    return 0;
}

static int // 0:normal phase    1:recovery phase
recover_from_spin(Player *p)
{
    p->cnt--;
    p->dir += 30;
    if (p->cnt != 0) return 1; // yet in recovery phase
    p->status = P_NORMAL;
    p->dir = 0;
    p->vx = 0;
    p->vy = -VEL;
    return 0;
}

static int // 0:normal phase    1:recovery phase
recover_from_fuelempty(Player *p)
{
    p->cnt--;
    if (p->cnt != 0) return 1; // yet in recovery phase
    if (p->nlife < 0) {
	p->status = P_RETIRE;
	pack_playerinfo2();
	pack_gameretire(p);
    }
    else {
	p->status = P_NORMAL;
	init_player(p2id(p), 0);
    }
    return 0;
}

static void
move_player(void)
{
    int i;
    Player *p;

    for (i = 0; i < Nclient + Nenemy; i++) {
	p = P + i;

	switch (p->status) {
	  case P_CRASH:
	    recover_from_crash(p);
	    break;
	  case P_SPIN:
	    recover_from_spin(p);
	    break;
	  case P_FUELEMPTY:
	    recover_from_fuelempty(p);
	    break;
	  case P_RETIRE:
	  case P_NOTLOGGEDIN:
	    move_position4retired(p);
	    break;
	  default:
	    if (i >= Nclient) {
		move_enemy_position(p);
	    }
	    else if (p->ctrlstyle == P_CLASSIC) {
		move_position_classic(p);
	    }
	    else {
		move_position_modern(p);
		move_direction_modern(p);
	    }
	    generate_smoke(p);
	    break;
	}
    }
}

static void
check_finish(void)
{
    int i;
    Player *p;

    if (NflagLeft == 0) {
	for (i = 0; i < Nclient; i++) {
	    p = P + i;
	    if (p->status == P_RETIRE ||
		p->status == P_NOTLOGGEDIN) continue;
	    p->status = P_FINISH;
	}
	pack_playerinfo2();
	pack_gamefinish();
	Gs = GAME_FINISH;
	GsCnt = 0;
    }
}

static void
check_over(void)
{
    int i;
    Player *p;
    int gameover = 1;

    for (i = 0; i < Nclient; i++) {
	p = P + i;
	if (p->status != P_RETIRE && p->status != P_NOTLOGGEDIN) {
	    gameover = 0;
	}
    }
    if (gameover) {
	Gs = GAME_OVER;
	GsCnt = 0;
	pack_gameover();
    }
}

static void
check_fuel(void)
{
    int i;
    Player *p;

    for (i = 0; i < Nclient; i++) {
	p = P + i;
	if (p->fuel > 0 ||
	    p->status == P_FUELEMPTY ||
	    p->status == P_RETIRE ||
	    p->status == P_NOTLOGGEDIN) continue;

	p->status = P_FUELEMPTY;
	p->cnt = 100;
	p->nlife--;
    }
}

static void
check_hit(void)
{
    int i, j;
    int mx, my; // player position on the map
    int obj; // object right at the car position
    Player *p;

    for (i = 0; i < Nclient + Nenemy; i++) {
	p = P + i;
	if (p->status == P_CRASH ||
	    p->status == P_RETIRE ||
	    p->status == P_NOTLOGGEDIN) continue;

	mx = PF2M_SCALE(p->x);
	my = PF2M_SCALE(p->y);
	obj = Map[my][mx];

	//	if (i == 1) printf("mx:%d  my:%d  obj:%d\n", mx, my, obj);

	if (i < Nclient) {
	    switch (obj) {
	      case MAP_LUCKY:
		Map[my][mx] = MAP_NONE;
		p->score += 1000 * p->fuel / MAX_FUEL;
		p->fuel = MAX_FUEL;
		p->nflag++;
		p->score += p->flagscore * p->special;
		p->flagscore += 100;
		NflagLeft--;
		pack_map();
		pack_stageinfo();
		pack_playerinfo2();
		if (p->score > Highscore) {
		    Highscore = p->score;
		    pack_highscore(Highscore);
		}
		break;
	      case MAP_SPECIAL:
		Map[my][mx] = MAP_NONE;
		p->nflag++;
		p->special = 2;
		p->score += p->flagscore * p->special;
		p->flagscore += 100;
		NflagLeft--;
		pack_map();
		pack_stageinfo();
		pack_playerinfo2();
		if (p->score > Highscore) {
		    Highscore = p->score;
		    pack_highscore(Highscore);
		}
		break;
	      case MAP_FLAG:
		Map[my][mx] = MAP_NONE;
		p->nflag++;
		p->score += p->flagscore * p->special;
		p->flagscore += 100;
		NflagLeft--;
		pack_map();
		pack_stageinfo();
		pack_playerinfo2();
		if (p->score > Highscore) {
		    Highscore = p->score;
		    pack_highscore(Highscore);
		}
		break;
	      case MAP_ROCK:
		Map[my][mx] = MAP_NONE;
		crash_car(p);
		pack_map();
		pack_playerinfo2();
		break;
	      default:
		break;
	    }
	}
	// check position of other cars
	for (j = 0; j < Nclient + Nenemy; j++) {
	    int dx, dy;
	    Player *q = P + j;
	    if (q == p) continue; // skip myself
	    if (q->status == P_RETIRE || 
		q->status == P_NOTLOGGEDIN) continue; // skip retired player
	    dx = q->x - p->x;
	    dy = q->y - p->y;
	    if (abs(dx) > RES/2 || abs(dy) > RES/2) continue;
	    if ((i >= Nclient && p->status == P_NORMAL && p->cnt > 0) ||
		(j >= Nclient && q->status == P_NORMAL && q->cnt > 0)) continue;
	    crash_car(p);
	    crash_car(q);
	}

	// check position of smokes
	for (j = 0; j < MAX_SMOKE; j++) {
	    int dx, dy;
	    if (Smoke[j][2] == 0) continue;
	    dx = Smoke[j][0] - p->x;
	    dy = Smoke[j][1] - p->y;
	    if (abs(dx) > RES/4 || abs(dy) > RES/4) continue;
	    spin_car(p);
	}
    }
}

static void
crash_car(Player *p)
{
    p->status = P_CRASH;
    p->cnt = 100;
    if (p2id(p) < Nclient) {
	p->nlife--;
    }
}

static void
spin_car(Player *p)
{
    if (p->status == P_NORMAL) {
	p->status = P_SPIN;
	p->cnt = 50;
    }
}

static void
generate_smoke(Player *p)
{
    int i, cnt;
    int key = p->key;

    if (p->status == P_RETIRE ||
	p->status == P_NOTLOGGEDIN) return;

    if (key & MASK_SPACEKEY) {
	for (i = 0, cnt = 1; i < MAX_SMOKE && cnt < MAX_TRAIL; i++) {
	    if (Smoke[i][2]) continue;

	    Smoke[i][0] = p->trail[cnt][0];
	    Smoke[i][1] = p->trail[cnt][1];
	    Smoke[i][2] = 100;
	    cnt++;
	}
	if (cnt) pack_smoke();
	p->fuel -= SMOKE_FUEL;
	p->key &= ~MASK_SPACEKEY;
    }
}

// decay existing smokes
static void
decay_smoke(void)
{
    int i, cnt;

    for (i = 0, cnt = 0; i < MAX_SMOKE; i++) {
	if (Smoke[i][2] == 0) continue;
	Smoke[i][2]--;
	if (Smoke[i][2] == 0) cnt++;
    }
    if (cnt) pack_smoke();
}

static void
move_position4retired(Player *p)
{
    int key = p->key;
    int v = 50;

    // position
    if (key & MASK_LEFTKEY) {
	p->x -= v;
    }
    if (key & MASK_RIGHTKEY) {
	p->x += v;
    }
    if (key & MASK_UPKEY) {
	p->y -= v;
    }
    if (key & MASK_DOWNKEY) {
	p->y += v;
    }
    limit_range(&p->x, 0, PF_WIDTH-1);
    limit_range(&p->y, 0, PF_HEIGHT-1);
}

static void
move_position_classic(Player *p)
{
    int i;
    int key;
    int mx, my, mx0, my0, mvx, mvy;
    int x0, y0;

    p->fuel--;

    key = p->key;
    mx = PF2M_SCALE(p->x);
    my = PF2M_SCALE(p->y);
    x0 = M2PF_SCALE(mx);
    y0 = M2PF_SCALE(my);

    if (p->vy > 0) p->dir = 180;
    if (p->vy < 0) p->dir = 0;
    if (p->vx > 0) p->dir = -90;
    if (p->vx < 0) p->dir = 90;
    if (p->vy && (p->y == y0)) { // moving vertically
	mvy = (p->vy > 0) ? +1 : -1;
	if ((key & MASK_RIGHTKEY) &&
	    mx < MAP_WIDTH - 1 && Map[my][mx + 1] != MAP_WALL) {
	    p->dir -= 45;
	    p->vx = +VEL;
	    p->vy = 0;
	}
	else if ((key & MASK_LEFTKEY) &&
		 mx > 0 && Map[my][mx - 1] != MAP_WALL) {
	    p->dir += 45;
	    p->vx = -VEL;
	    p->vy = 0;
	}
	else if (((key & MASK_DOWNKEY) && mvy < 0) ||
		 ((key & MASK_UPKEY) && mvy > 0)) {
	    p->dir += -90;
	    p->vy *= -1;
	}
	else if (my + mvy < 0 || my + mvy >= MAP_HEIGHT ||
		 Map[my + mvy][mx] == MAP_WALL) {
	    if (Map[my][mx - mvy] == MAP_WALL ||
		mx - mvy < 0 || mx - mvy >= MAP_WIDTH) {
		p->dir += 90;
		p->vy *= -1;
	    }
	    else {
		p->dir -= 45;
		p->vx = - p->vy;
		p->vy = 0;
	    }
	}
    }
    else if (p->vx && (p->x == x0)) { // moving horizontally
	mvx = (p->vx > 0) ? +1 : -1;
	if ((key & MASK_DOWNKEY) &&
	    my < MAP_HEIGHT - 1 && Map[my + 1][mx] != MAP_WALL) {
	    p->dir += 45;
	    p->vx = 0;
	    p->vy = +VEL;
	}
	else if ((key & MASK_UPKEY) &&
		 my > 0 && Map[my - 1][mx] != MAP_WALL) {
	    p->dir -= 45;
	    p->vx = 0;
	    p->vy = -VEL;
	}
	else if (((key & MASK_RIGHTKEY) && mvx < 0) ||
		 ((key & MASK_LEFTKEY) && mvx > 0)) {
	    p->dir += 90;
	    p->vx *= -1;
	}
	else if (mx + mvx < 0 || mx + mvx >= MAP_WIDTH ||
		 Map[my][mx + mvx] == MAP_WALL) {
	    if (Map[my + mvx][mx] == MAP_WALL ||
		my + mvx < 0 || my + mvx >= MAP_HEIGHT) {
		p->dir += 90;
		p->vx *= -1;
	    }
	    else {
		p->dir -= 45;
		p->vy = p->vx;
		p->vx = 0;
	    }
	}
    }

    // update car position
    p->x += p->vx;
    p->y += p->vy;

    // update trail position when the car moved to the next cell
    mx0 = PF2M_SCALE(p->trail[0][0]);
    my0 = PF2M_SCALE(p->trail[0][1]);
    mx = PF2M_SCALE(p->x);
    my = PF2M_SCALE(p->y);
    if (mx0 != mx || my0 != my) { // car moved to next cell
	for (i = MAX_TRAIL-1; i >= 1; i--) {
	    p->trail[i][0] = p->trail[i-1][0];
	    p->trail[i][1] = p->trail[i-1][1];
	}
	p->trail[0][0] = p->x;
	p->trail[0][1] = p->y;
    }
}

static void
move_enemy_position(Player *p)
{
    int i;
    int key;
    int mx, my, mx0, my0, mvx, mvy;
    int x0, y0;

    if (p->cnt > 0) { // not ready to move
	p->cnt--;
	return;
    }
    key = p->key;
    mx = PF2M_SCALE(p->x);
    my = PF2M_SCALE(p->y);
    x0 = M2PF_SCALE(mx);
    y0 = M2PF_SCALE(my);

    if (p->vy > 0) p->dir = 180;
    if (p->vy < 0) p->dir = 0;
    if (p->vx > 0) p->dir = -90;
    if (p->vx < 0) p->dir = 90;
    if (p->vy && (p->y == y0)) { // moving vertically
	mvy = (p->vy > 0) ? +1 : -1;
	if ((key & MASK_RIGHTKEY) &&
	    mx < MAP_WIDTH - 1 && Map[my][mx + 1] != MAP_WALL && Map[my][mx + 1] != MAP_ROCK) {
	    p->dir -= 45;
	    p->vx = +VEL;
	    p->vy = 0;
	}
	else if ((key & MASK_LEFTKEY) &&
		 mx > 0 && Map[my][mx - 1] != MAP_WALL && Map[my][mx - 1] != MAP_ROCK) {
	    p->dir += 45;
	    p->vx = -VEL;
	    p->vy = 0;
	}
	else if (((key & MASK_DOWNKEY) && mvy < 0) ||
		 ((key & MASK_UPKEY) && mvy > 0)) {
	    p->dir += -90;
	    p->vy *= -1;
	}
	else if (my + mvy < 0 || my + mvy >= MAP_HEIGHT ||
		 Map[my + mvy][mx] == MAP_WALL ||
		 Map[my + mvy][mx] == MAP_ROCK) {
	    if (Map[my][mx - mvy] == MAP_WALL ||
		Map[my][mx - mvy] == MAP_ROCK ||
		mx - mvy < 0 || mx - mvy >= MAP_WIDTH) {
		p->dir += 90;
		p->vy *= -1;
	    }
	    else {
		p->dir -= 45;
		p->vx = - p->vy;
		p->vy = 0;
	    }
	}
    }
    else if (p->vx && (p->x == x0)) { // moving horizontally
	mvx = (p->vx > 0) ? +1 : -1;
	if ((key & MASK_DOWNKEY) &&
	    my < MAP_HEIGHT - 1 && Map[my + 1][mx] != MAP_WALL && Map[my + 1][mx] != MAP_ROCK) {
	    p->dir += 45;
	    p->vx = 0;
	    p->vy = +VEL;
	}
	else if ((key & MASK_UPKEY) &&
		 my > 0 && Map[my - 1][mx] != MAP_WALL && Map[my - 1][mx] != MAP_ROCK) {
	    p->dir -= 45;
	    p->vx = 0;
	    p->vy = -VEL;
	}
	else if (((key & MASK_RIGHTKEY) && mvx < 0) ||
		 ((key & MASK_LEFTKEY) && mvx > 0)) {
	    p->dir += 90;
	    p->vx *= -1;
	}
	else if (mx + mvx < 0 || mx + mvx >= MAP_WIDTH ||
		 Map[my][mx + mvx] == MAP_WALL ||
		 Map[my][mx + mvx] == MAP_ROCK) {
	    if (Map[my + mvx][mx] == MAP_WALL ||
		Map[my + mvx][mx] == MAP_ROCK ||
		my + mvx < 0 || my + mvx >= MAP_HEIGHT) {
		p->dir += 90;
		p->vx *= -1;
	    }
	    else {
		p->dir -= 45;
		p->vy = p->vx;
		p->vx = 0;
	    }
	}
    }

    // update car position
    p->x += p->vx;
    p->y += p->vy;

    // update trail position when the car moved to the next cell
    mx0 = PF2M_SCALE(p->trail[0][0]);
    my0 = PF2M_SCALE(p->trail[0][1]);
    mx = PF2M_SCALE(p->x);
    my = PF2M_SCALE(p->y);
    if (mx0 != mx || my0 != my) { // car moved to next cell
	for (i = MAX_TRAIL-1; i >= 1; i--) {
	    p->trail[i][0] = p->trail[i-1][0];
	    p->trail[i][1] = p->trail[i-1][1];
	}
	p->trail[0][0] = p->x;
	p->trail[0][1] = p->y;
    }
}

static void
move_position_modern(Player *p)
{
    int i;
    int key = p->key;
    int da = 6;
    int vmax = 20;
    int mu = 3;
    int mx, my, mx0, my0; // player position on map

    // acceleration
    p->ax = p->ay = 0;
    if (key & MASK_LEFTKEY) {
	p->ax = -da;
    }
    if (key & MASK_RIGHTKEY) {
	p->ax = +da;
    }
    if (key & MASK_UPKEY) {
	p->ay = -da;
    }
    if (key & MASK_DOWNKEY) {
	p->ay = +da;
    }
    if (p->ax || p->ay) p->fuel--;

    // velocity
    if (p->vx > 0) { // friction
	p->vx -= mu;
    }
    else if (p->vx < 0) {
	p->vx += mu;
    }
    p->vx += p->ax;
    limit_range(&p->vx, -vmax, +vmax);
    if (p->vy > 0) { // friction
	p->vy -= mu;
    }
    else if (p->vy < 0) {
	p->vy += mu;
    }
    p->vy += p->ay;
    limit_range(&p->vy, -vmax, +vmax);

    mx = PF2M_SCALE(p->x + p->vx);
    my = PF2M_SCALE(p->y);
    limit_range(&mx, 0, MAP_WIDTH-1);
    if (Map[my][mx] == MAP_WALL) { // reflection
	p->vx *= -1;
    }
    mx = PF2M_SCALE(p->x + p->vx);
    my = PF2M_SCALE(p->y + p->vy);
    limit_range(&mx, 0, MAP_WIDTH-1);
    limit_range(&my, 0, MAP_HEIGHT-1);
    if (Map[my][mx] == MAP_WALL) { // reflection
	p->vy *= -1;
    }

    // position
    p->x += p->vx;
    p->y += p->vy;
    limit_range(&p->x, 0, PF_WIDTH-1);
    limit_range(&p->y, 0, PF_HEIGHT-1);

    //    fprintf(stderr, "vmax:%d  vx:%d  vy:%d\n", vmax, p->vx, p->vy);
    // save the latest three trails
    
    mx0 = PF2M_SCALE(p->trail[0][0]);
    my0 = PF2M_SCALE(p->trail[0][1]);
    mx = PF2M_SCALE(p->x);
    my = PF2M_SCALE(p->y);
    if (mx0 != mx || my0 != my) { // car moved to next cell
	for (i = MAX_TRAIL-1; i >= 1; i--) {
	    p->trail[i][0] = p->trail[i-1][0];
	    p->trail[i][1] = p->trail[i-1][1];
	}
	p->trail[0][0] = p->x;
	p->trail[0][1] = p->y;
    }
}

static void
move_direction_modern(Player *p)
{
    int key;
    key = p->key;

    // direction
    if (key & MASK_UPKEY) {
	p->dir = 0;
    }
    if (key & MASK_DOWNKEY) {
	p->dir = 180;
    }
    if (key & MASK_LEFTKEY) {
	if (key & MASK_UPKEY) {
	    p->dir = 45;
	}
	else if (key & MASK_DOWNKEY) {
	    p->dir = 135;
	}
	else {
	    p->dir = 90;
	}
    }
    if (key & MASK_RIGHTKEY) {
	if (key & MASK_UPKEY) {
	    p->dir = -45;
	}
	else if (key & MASK_DOWNKEY) {
	    p->dir = -135;
	}
	else {
	    p->dir = -90;
	}
    }
}

static void
myDisplay(void)
{
    int i;
    char buf[128];
    double h = 2.5 / (Nclient + 2);

    glClear(GL_COLOR_BUFFER_BIT);
    glColor3d(0.0, 1.0, 0.0);
    sprintf(buf, "ID      x          y           key           name");
    putString(buf, -0.9, 1.0 - 0.7 * h, 2);
    glColor3d(1.0, 1.0, 0.0);
    for (i = 0; i < Nclient; i++) {
	sprintf(buf, "%02d    %+05d    %+05d    0x%03x        %s",
		i, P[i].x, P[i].y, P[i].key, P[i].name);
	putString(buf, -0.9, 1.0 - (1.5 + i) * h, 2);
    }
    glutSwapBuffers();
}

static void
myReshape(int w, int h)
{
    w = WIN_WIDTH;
    h = 20 + 20 * Nclient;
    glutReshapeWindow(w, h);
    glViewport(0, 0, w, h);
}

static void
myTimer(int id)
{
    int i;

    for (i = 0; i < Nclient; i++) {
	Player *p = P + i;
	if (p->key & MASK_QKEY) {
	    pack_gamequit();
	    p->key &= ~MASK_QKEY;
	    sleep(2);
	    die(0, Nclient, Soc);
	}
	if (p->key & MASK_SKEY) {
	    p->key &= ~MASK_SKEY;
	    pack_gamepause();
	    toggle_pause();
	}
    }
    (game_func[Gs])();
    glutPostRedisplay();
    glutTimerFunc(TimerPeriod, myTimer, 0);
}

static void
game_title(void)
{
    int i, j;
    int onesec = TimerPeriod;
    static int startcnt = 0;

    recv_from_clients();

    if (GsCnt == 0) { // initialize title
	pack_gametitle(-1);
	for (i = 0; i < Nclient + Nenemy; i++) {
	    init_player(i, 2);
	}
	pack_playerinfo2();
	Stage = 1;
	Nenemy = 1;
	TimerPeriod = INIT_TIMER_PERIOD;
	GsCnt = 1; // initialization done
    }
    for (i = 0; i < Nclient; i++) {
	Player *p = P + i;
	if (p->key & MASK_LEFTKEY) {
	    p->ctrlstyle = P_CLASSIC;
	    pack_playerinfo2();
	}
	else if (p->key & MASK_RIGHTKEY) {
	    p->ctrlstyle = P_MODERN;
	    pack_playerinfo2();
	}
    }
    for (i = 0, j = 0; i < Nclient; i++) {
	Player *p = P + i;
	if (p->key & MASK_SPACEKEY) {
	    p->status = P_NORMAL;
	    pack_playerinfo2();
	    GsCnt = 2; // begin counting down
	    startcnt = 0;
	}
	if (p->status == P_NORMAL) j++;
    }
    if (GsCnt != 2) return;

    if (startcnt % onesec == 0) {
	pack_gametitle(10 - startcnt / onesec);
	//	fprintf(stderr, "pack_gametitle %d\n", 10 - startcnt / onesec);
    }
    startcnt++;

    // start immediately, if all players logged in or wait time expired.
    if (startcnt == onesec * 10 || j == Nclient) {
	for (i = 0, j = 0; i < Nclient; i++) {
	    Player *p = P + i;
	    if (p->status == P_NOTLOGGEDIN) {
		p->x = PF_WIDTH/2;
		p->y = PF_HEIGHT/2;
	    }
	    p->score = 0;
	}
	pack_playerinfo();
	pack_gametitle(-2); // start immediately
	Gs = GAME_READY;
	GsCnt = 0;
	return;
    }
}

static void
game_finish(void)
{
    GsCnt++;
    if (GsCnt == 50) {
	Gs = GAME_RESULT;
	GsCnt = 0;
    }
}

static void
game_over(void)
{
    int i;

    if (GsCnt == 0) {
	for (i = 0; i < Nclient; i++) {
	    Player *p = P + i;
	    p->ntotalflag += p->nflag;
	}
    }
    GsCnt++;
    if (GsCnt == 50) {
	Gs = GAME_FINALRESULT;
	GsCnt = 0;
    }
}

static void
game_result(void)
{
    int i;
    int winner = 0;

    if (GsCnt == 0) {
	for (i = 0; i < Nclient; i++) {
	    Player *p = P + i;
	    if (p->status == P_RETIRE ||
		p->status == P_NOTLOGGEDIN) continue;
	    if (P[winner].nflag < p->nflag) {
		winner = p2id(p);
	    }
	    p->ntotalflag += p->nflag;
	}
	P[winner].nwin++;
	pack_playerinfo2();
	pack_gameresult(winner);
    }
    GsCnt++;
    if (GsCnt == 200) {
	Stage++;
	TimerPeriod = TimerPeriod * 9 / 10;
	Nenemy = Stage;
	if (Nenemy + Nclient > MAX_CLIENT) {
	    Nenemy = MAX_CLIENT - Nclient;
	}
	Gs = GAME_READY;
	GsCnt = 0;
    }
}

static void
game_finalresult(void)
{
    int winner = 0;

    pack_playerinfo2();
    pack_gamefinalresult();

    GsCnt++;
    if (GsCnt == 200) {
	Gs = GAME_TITLE;
	GsCnt = 0;
    }
}

static void
game_ready(void)
{
    int ret = 0;
    int dt = 30;
    int i;

    if (GsCnt == 0) {
	pack_gameready(0);
	for (i = 0; i < MAX_SMOKE; i++) {
	    Smoke[i][2] = 0;
	}
	pack_smoke();
	load_map(Stage);
	pack_map();
	pack_stageinfo();
	for (i = 0; i < Nclient + Nenemy; i++) {
	    init_player(i, 1);
	}
	pack_playerinfo2();
    }
    GsCnt++;

    if (GsCnt == 1 * dt) {
	pack_gameready(2);
    }
    if (GsCnt == 2 * dt) {
	pack_gameready(1);
    }
    if (GsCnt == 3 * dt) {
	pack_gamestart(Stage);
	Gs = GAME_ONPLAY;
	GsCnt = 0;
    }
}

static void
game_onplay(void)
{
    recv_from_clients();
    control_enemy();
    move_player();
    check_hit();
    check_fuel();
    check_finish();
    check_over();
    decay_smoke();
    pack_playerinfo();
}

static void
game_pause(void)
{
    int i;
    recv_from_clients();
    for (i = 0; i < Nclient; i++) {
	Player *p = P + i;
	if (p->status == P_NOTLOGGEDIN ||
	    p->status == P_RETIRE) {
	    move_position4retired(p);
	}
    }
}

static void
load_map(int stage)
{
    int i, j, id;
    char val;
    char sbuf[2] = {0, 0};
    char buf[256];
    FILE *fp;
    char *fname[] = {
	"data/map%d",
	"../data/map%d",
	"../../data/map%d",
    };

    for (i = 0; i < sizeof(fname)/sizeof(fname[0]); i++) {
	sprintf(buf, fname[i], (stage - 1) % (MAX_STAGE) + 1);
	fp = fopen(buf, "r");
	if (fp) break;
    }
    if (!fp) {
	perror("fopen in load_map()");
	die(1, Nclient, Soc);
    }
    Nflag = 0;
    for (i = 0; i < MAP_HEIGHT; i++) {
	fgets(buf, sizeof(buf), fp);
	for (j = 0; j < MAP_WIDTH; j++) {
	    if (isdigit(buf[j])) {
		sbuf[0] = buf[j];
		id = atoi(sbuf);
		if (id >= 0 && id < Nclient) {
		    Player *p = P + id;
		    p->initx = M2PF_SCALE(j);
		    p->inity = M2PF_SCALE(i);
		    buf[j] = MAP_NONE;
		}
		if (id >= MAX_CLIENT - Nenemy && id < MAX_CLIENT) {
		    Player *p = P + MAX_CLIENT - id - 1 + Nclient;
		    p->initx = M2PF_SCALE(j);
		    p->inity = M2PF_SCALE(i);
		    buf[j] = MAP_NONE;
		}
	    }
	    switch (buf[j]) {
	      case ' ':
		val = MAP_NONE;
		break;
	      case '*':
		val = MAP_WALL;
		break;
	      case 'f':
		val = MAP_FLAG;
		Nflag++;
		break;
	      case 'r':
		val = MAP_ROCK;
		break;
	      case 'l':
		val = MAP_LUCKY;
		Nflag++;
		break;
	      case 's':
		val = MAP_SPECIAL;
		Nflag++;
		break;
	      default:
		val = MAP_NONE;
		break;
	    }
	    Map[i][j] = val;
	}
    }
    fclose(fp);
    NflagLeft = Nflag;
    // print_map();
}

// just for debugging purpose
static void
print_map(void)
{
    int i, j;
    char c;

    for (i = 0; i < MAP_HEIGHT; i++) {
	printf("\n", c);
	for (j = 0; j < MAP_WIDTH; j++) {
	    switch (Map[i][j]) {
	      case MAP_NONE:
		c = '.';
		break;
	      case MAP_WALL:
		c = '*';
		break;
	      case MAP_FLAG:
		c = 'f';
		break;
	      case MAP_ROCK:
		c = 'r';
		break;
	      case MAP_LUCKY:
		c = 'l';
		break;
	      case MAP_SPECIAL:
		c = 's';
		break;
	      default:
		c = '?';
		break;
	    }
	    printf("%c", c);
	}
    }
    printf("\n", c);
}

static void
myKeyboard(unsigned char key, int x, int y)
{
    if (key == 'q') {
	pack_gamequit();
	sleep(2);
	die(0, Nclient, Soc);
    }
    if (key == 's') {
	pack_gamepause();
	toggle_pause();
    }
}

static void
toggle_pause(void)
{
    if (Gs == GAME_PAUSE) {
	Gs = GAME_ONPLAY;
    }
    else if (Gs == GAME_ONPLAY) {
	Gs = GAME_PAUSE;
    }
}

static void
control_enemy(void)
{
    int i, j;
    int dist, distmin;
    int dx, dy;
    int mx, my;
    Player *target; // cat to pursue

    for (i = Nclient; i < Nclient + Nenemy; i++) {
	Player *p = P + i; // enemy car

	for (j = 0; j < Nclient; j++) {
	    Player *q = P + j; // player car

	    if (q->status == P_NOTLOGGEDIN ||
		q->status == P_RETIRE) continue;
	    dist = abs(q->x - p->x) + abs(q->y - p->y);
	    if (j == 0) {
		distmin = dist;
		target = q;
	    }
	    else if (distmin > dist) {
		distmin = dist;
		target = q;
	    }
	}
	dx = target->x - p->x;
	dy = target->y - p->y;

	p->key = 0;
	mx = PF2M_SCALE(p->x);
	my = PF2M_SCALE(p->y);

	if (abs(dx) > abs(dy)) { // horizontal adjust
	    if (dx > 0 && Map[my][mx + 1] == MAP_NONE) {
		p->key |= MASK_RIGHTKEY;
	    }
	    else if (dx < 0 && Map[my][mx - 1] == MAP_NONE) {
		p->key |= MASK_LEFTKEY;
	    }
	    else if (dy > 0 && Map[my + 1][mx] == MAP_NONE) {
		p->key |= MASK_DOWNKEY;
	    }
	    else if (dy < 0 && Map[my - 1][mx] == MAP_NONE) {
		p->key |= MASK_UPKEY;
	    }
	}
	else { // vertical adjust
	    if (dy > 0 && Map[my + 1][mx] == MAP_NONE) {
		p->key |= MASK_DOWNKEY;
	    }
	    else if (dy < 0 && Map[my - 1][mx] == MAP_NONE) {
		p->key |= MASK_UPKEY;
	    }
	    else if (dx > 0 && Map[my][mx + 1] == MAP_NONE) {
		p->key |= MASK_RIGHTKEY;
	    }
	    else if (dx < 0 && Map[my][mx - 1] == MAP_NONE) {
		p->key |= MASK_LEFTKEY;
	    }
	}

#if 0
	{
	    char buf[32];
	    if (p->key & MASK_LEFTKEY)  sprintf(buf, "left");
	    if (p->key & MASK_RIGHTKEY) sprintf(buf, "right");
	    if (p->key & MASK_UPKEY)    sprintf(buf, "up");
	    if (p->key & MASK_DOWNKEY)  sprintf(buf, "down");
	    fprintf(stderr, "i:%d  key:%s\n", p2id(p), buf);
	}
#endif
    }
}
